#include <stdio.h>
#include <tcl.h>
#include <termios.h>
#include <errno.h>
#include <sys/fcntl.h>

#include <sys/signal.h>
extern int Debug;

#define bitcode(X)            (1 << (16-X))

typedef struct _XTenSession {
	int	sessid;				/* session id */
	char name[6];			/* session handle */
	int fd;					/* the open tty */
	char devicename[30];	/* X10 /dev device name */
	Tcl_Command token;		/* to tell one session from the next */
	Tcl_Interp *interp;		/* Tcl Interpreter */
	Tcl_File tclfile;		/* Tcl file */
	Tcl_DString command;	/* dynamic command to execute */
	Tcl_DString args;		/* dynamic args to pass as client data */
} XTenSession;
	
static XTenSession *Sesslist[99];
static Tcl_HashTable aliasTable;

/*
 * Routing to handle receipt of async data from CP290
 * (Say, as the result of somebody pressing a switch on a powerhouse
 * mini controller or something like that - we need a way to update
 * the GUI properly in this event (if applicable). There's a callback
 * via command that we can execute from here to do so.
 */
extern char code2zone[];

void X10_Incoming(ClientData clientdata, int mask) 
{
	XTenSession *session = (XTenSession *) clientdata;
	unsigned char buf[13];
	char *cval;
	char zonecode;
	char basezone;
	char devices[80];
	Tcl_DString cmd_n_args;
	int function;
	int i;
	char *status[4] = { "{}", "{}", "on", "off"};
	char tcmd[10];

	devices[0] = (char) NULL;
	puts("Got read data!");
	fflush(stdout);
	read(session->fd, &buf[1], 12);
	zonecode = code2zone[buf[8] >> 4];
	basezone = code2zone[buf[11] >> 4];
	function = buf[8] & 0x0f;
	if (function == 15)
		return;	/* ignore dim code */

	for (i = 0; i < 8; i++) {
		if ((int) (buf[9] & (1 << i)) > 0) 
			sprintf(devices+strlen(devices), "%d ", 16-i);
	}
	for (i = 0; i < 8; i++) {
		if ((int) (buf[10] & (1 << i)) > 0) 
			sprintf(devices+strlen(devices), "%d ", 8-i);
	}
	/* get rid of trailing white space */
	devices[strlen(devices)-1] = (char) NULL;

	if (Debug) {
		printf("raw, %2.2x %2.2x\n", buf[9], buf[10]);
		printf("Code %c/%c, Devices %s, function %d\n", zonecode, 
				basezone, devices, function);
		fflush(stdout);
	}
	sprintf(tcmd, " %c {%s} %s ", zonecode, devices, status[function]);
	if ((cval = Tcl_DStringValue(&session->command)) != NULL) {
		Tcl_DStringInit(&cmd_n_args);
		Tcl_DStringAppend(&cmd_n_args, cval, strlen(cval));
		Tcl_DStringAppend(&cmd_n_args, tcmd, strlen(tcmd));
		Tcl_DStringAppend(&cmd_n_args, Tcl_DStringValue(&session->args), -1);
		Tcl_Eval(session->interp, Tcl_DStringValue(&cmd_n_args));

		Tcl_DStringFree(&cmd_n_args);
	}
	
}

/* 
 * Utility routine to setup X10 device/file/channel
 */

int X10_setupDevice(XTenSession *session, char *name) 
{
	Tcl_File tfd;
	int spoon=1;
	
	strncpy(session->devicename, name, 29);
	session->devicename[30] = (char) NULL;
	if ((session->fd = open(session->devicename, O_RDWR|O_EXCL)) < 0) {
		Tcl_AppendResult(session->interp, 
				"Bad tty name or permission problem: ", 
				session->devicename, (char *) NULL);
		return TCL_ERROR;
	}
	/* initialize port speed and other characteristics */
	if (set_tty(session->fd) < 0)	{
		close(session->fd);
		Tcl_AppendResult(session->interp, 
				"Could not communicate with interface.  Check your serial cables, battery, and power.", (char *) NULL);
		return TCL_ERROR;
	}

	x10_set_basezone(session->fd, 'a');
	x10_ack(session->fd);
	tfd = Tcl_GetFile((ClientData) session->fd, TCL_UNIX_FD);
	session->tclfile = tfd;
	Tcl_CreateFileHandler(tfd, TCL_READABLE, X10_Incoming,(ClientData) session);

	return TCL_OK;
}

/*
 * X10 session handler
 * This treats each handle as a command
 * if device was not specified and init was not used,
 * Then the configure command here must be called to initialize the 
 * serial device (cp290). Configure can also be used to change a
 * device or serial port.
 *
 * info returns the session and the corresponding device
 * 
 * otherwise we expect an alias or a code pair followed by an on or off
 * command to control the device
 */
static int X10_Command(ClientData clientData, Tcl_Interp *interp, 
	int argc, char **argv)
{
	XTenSession *session = (XTenSession *) clientData;
	int Listargc;
	char **Listargv;
	int i;
	int zone;
	int codes = 0;
	int action;
	int tcode;
	Tcl_HashEntry *entryPtr;

	if (strcmp(argv[1], "info") == 0) {
		Tcl_AppendElement(interp, session->name);
		Tcl_AppendElement(interp, session->devicename);
		return TCL_OK;
	} else if (strcmp(argv[1], "configure") == 0) {
		if (argc < 3) {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
				argv[0], " configure option value\"", (char *) NULL);
			return TCL_ERROR;
		}
		
		if (strcmp (argv[2], "-device") == 0) {
			if (argc == 3) {
				Tcl_AppendResult(interp, session->devicename, (char *) NULL);
				return TCL_OK;
			}
			if (argc != 4) {
				Tcl_AppendResult(interp, "bad args should be \"", argv[0],
						" configure -device <devname>\"", (char *) NULL);
				return TCL_ERROR;
			}
			if (session->fd != 0)		/* already set, reset it */
				close(session->fd);
			if (X10_setupDevice(session, argv[3]) != TCL_OK)
				return TCL_ERROR;
		} else if (strcmp(argv[2], "-basezone") == 0) {
			/* Set the base code for the X10 interface for the rocker switches*/
			if (argc != 4 || (argv[3][0] < 'a' || argv[3][0] > 'p')) {
				Tcl_AppendResult(interp, "bad args: should be \"",
						argv[0], " configure -basezone <a-p>\"", (char *) NULL);
				return TCL_ERROR;
			}
			x10_set_basezone(session->fd, argv[3][0]);
			x10_ack(session->fd);
		} else if (strcmp(argv[2], "-command") == 0) {
			/* enable a Tcl procedure to call on receipt of rocker switch
			   async notification */
			int i;

			if (argc == 3) {
				Tcl_AppendResult(interp, Tcl_DStringValue(&session->command),
						(char *) NULL);
				Tcl_AppendResult(interp, " ", Tcl_DStringValue(&session->args),
						(char *) NULL);
				return TCL_OK;
			}
			Tcl_DStringFree(&session->command);
			Tcl_DStringFree(&session->args);
			/* Collect arguments. 1st one is command, pass rest in as
			 * 'args' (ie variable length argument list for user to access)
			 */
			Tcl_DStringAppendElement(&session->command, argv[3]);
			if (argc > 4)
				for (i = 4; i < argc; i++) {
					Tcl_DStringAppendElement(&session->args, argv[i]);
				}
			return TCL_OK;
		} else {
			Tcl_AppendResult(session->interp, "bad option: \"", argv[2], 
					"\": should be -device, or -basezone", (char *) NULL);
			return TCL_ERROR;
		}
		return TCL_OK;
	}
		
	/* Has the session been configured */
	if (session->devicename[0] == (char) NULL) {
		Tcl_AppendResult(interp, 
				"No device configured for this session.\nUse: ",
				argv[0], " configure -device /dev/x10device", (char *) NULL);
		return TCL_ERROR;
	}

	/* Session called with alias name ? */
	if (argc == 3) {
		entryPtr = Tcl_FindHashEntry(&aliasTable, argv[1]);
		if (entryPtr == NULL) {
			Tcl_AppendResult(interp, "alias \"", argv[1], 
					"\" not currently defined.", (char *) NULL);
			return TCL_ERROR;
		}
		if (Tcl_SplitList(interp, Tcl_GetHashValue(entryPtr), 
					&Listargc, &Listargv) != TCL_OK)
			return TCL_ERROR;
		argv[3] = argv[2];		/* move on|off */
		argv[1] = Listargv[0];	/* set first arg to zone name */
		argv[2] = Listargv[1];	/* set second arg to code list */
		argc++;
	}
		
	if (argc != 4) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], " zone code_list (on|off)\"", (char *) NULL);
		return TCL_ERROR;
	}

	/* Session references with code, or alias converted to code/device */
	zone = argv[1][0] - 'a';
	if (Tcl_SplitList(interp, argv[2], &Listargc, &Listargv) != TCL_OK)
		return TCL_ERROR;
	for (i = 0; i < Listargc; i++) {
		if (Tcl_GetInt(interp, Listargv[i], &tcode) != TCL_OK) {
			Tcl_AppendResult(interp, " while processing device codes",
					(char *) NULL);
			return TCL_ERROR;
		}
		codes |= bitcode(tcode);
	}

	if (strcmp(argv[3], "off") == 0) action = 3; 
	else action = 2;

	/* flush the device - don't need spurious inputs from the rocker
	 * switch interfering */

	x10_flush(session->fd);

	/* temporarily disable handler while we process the response */
	Tcl_CreateFileHandler(session->tclfile, 0, X10_Incoming, (ClientData) session);
	if (x10_send_command(session->fd, zone, codes, action)) {
		Tcl_AppendResult(interp, 
			"There has been an X10 signal transmission error, try again.", 
			(char *) NULL);
		return TCL_ERROR;
	}
	Tcl_CreateFileHandler(session->tclfile, TCL_READABLE, X10_Incoming, (ClientData) session);

	return TCL_OK;
}


static void X10Destroy(ClientData clientData)
{
	XTenSession *session = (XTenSession *) clientData;

/*	Tcl_UnregisterChannel(session->interp, session->channel);*/
	Tcl_DStringFree(&session->command);
	Tcl_DStringFree(&session->args);
	Tcl_DeleteFileHandler(session->tclfile);
	Tcl_FreeFile(session->tclfile);
	close(session->fd);
	Sesslist[session->sessid] = (XTenSession *) NULL;
	free(session);
}

/*
 * evaluated x10 command options and commands
 *
 * init [device]  - initialize a x10 session, with optional device
 * alias [session] [[zonecode devicecode]] - show or set aliases
 * list - show all session and device pairs for all extant x10 sessions
 */

int X10_Cmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
	static int id = 0;
	int i;
	XTenSession *session;
	Tcl_HashEntry *entryPtr;

	
	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"",
			argv[0], "\" option", (char *) NULL);
		return TCL_ERROR;
	}

	if (strcmp(argv[1], "init") == 0) {
		
		if (id > 99) {
			Tcl_AppendResult(interp, "Out of X10 session names. Why would you want more than 99 sessions?", (char *) NULL);
			return TCL_ERROR;
		}
				
		session = (XTenSession *) ckalloc(sizeof(XTenSession));
		memset((char *) session, 0, sizeof(XTenSession));
		session->sessid = id;
		session->interp = interp;
		Sesslist[id] = session;

		if (argc == 3) {		/* can pass device via init too */
			if (X10_setupDevice(session, argv[2]) != TCL_OK) {
				close(session->fd);
				free(session);
				Sesslist[id] = NULL;
				return TCL_ERROR;
			}
		}
		Tcl_DStringInit(&session->command);
		sprintf(session->name, "xten%d", id++);
		session->token = Tcl_CreateCommand(interp, session->name, X10_Command,
			(ClientData) session, X10Destroy);
		Tcl_SetResult(interp, session->name, TCL_STATIC);
		return TCL_OK;
	} else if (strcmp(argv[1], "list") == 0) {
		for (i = 0; i < 100; i++) {
			char buf[80];
			if (Sesslist[i] != NULL) {
				sprintf(buf, "%s %s", Sesslist[i]->name, Sesslist[i]->devicename);
				Tcl_AppendElement(interp, buf);
			}
		}
		return TCL_OK;
	} else if (strcmp(argv[1], "alias") == 0) {
		/* show all aliases */
		if (argc == 2) {
			Tcl_HashSearch search;
			entryPtr = Tcl_FirstHashEntry(&aliasTable, &search);
			while (entryPtr) {
				Tcl_AppendElement(interp,
						Tcl_GetHashKey(&aliasTable, entryPtr));
				entryPtr = Tcl_NextHashEntry(&search);
			}
			return TCL_OK;
		/* show a particular alias */
		} else if (argc == 3) {
			entryPtr = Tcl_FindHashEntry(&aliasTable, argv[2]);
			if (entryPtr) {
				Tcl_SetResult(interp, (char *) Tcl_GetHashValue(entryPtr),
						TCL_STATIC);
			}
			return TCL_OK;
		/* install a new alias */
		} else if (argc == 4) {
			int isNew;
			if (strlen(argv[3]) < 2) {
				Tcl_AppendResult(interp, 
					"Alias name must be longer than one character to distinguish from zone designator.", (char *) NULL);
				return TCL_ERROR;
			}
			entryPtr = Tcl_CreateHashEntry(&aliasTable, argv[2], &isNew);
			if (!isNew) {
				ckfree((char *) Tcl_GetHashValue(entryPtr));
			}
			if (*argv[3] == '\0') {
				Tcl_DeleteHashEntry(entryPtr);
			} else {
				Tcl_SetHashValue(entryPtr, strdup(argv[3]));
			}
			return TCL_OK;
		} else {
			Tcl_AppendResult(interp, "wrong # args: should be \"",
				argv[0], "\" alias aliasname {device zone code}", (char *) NULL);
			return TCL_ERROR;
		}
	}


	Tcl_AppendResult(interp, "bad option: \"", argv[1], 
		"\": should be init init or list", (char *) NULL);
		
	return TCL_ERROR;
}

/*
 * Start the ball rolling
 * Create the x10 command
 * provide the X10 package
 * Initialize the hash table containing all x10 sessions
 * evaluate the .x10rc file in the users home directory
 */
X10_Init(Tcl_Interp *interp) 
{
	char script[80];
	sprintf(script, "%s/.x10rc", getenv("HOME"));
	
	Tcl_CreateCommand(interp, "x10", X10_Cmd,
			(ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_PkgProvide(interp, "X10", "1.0");
	memset((char *) Sesslist, 0, sizeof(Sesslist));
	Tcl_InitHashTable(&aliasTable, TCL_STRING_KEYS);

	/* source init file */
	Tcl_EvalFile(interp, script);
	return TCL_OK;
}
